/* Copyright (C) 2006,2007 Jacco de Leeuw <jacco2@dds.nl>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.  See <http://www.fsf.org/copyleft/gpl.txt>.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 * 
 * This program is released under the GPL with the additional exemption that
 * compiling, linking, and/or using OpenSSL is allowed.
 *
 */

// p12imprtDlg.cpp : implementation file
//

#include "stdafx.h"
#include "p12imprt.h"
#include "p12imprtDlg.h"

/* Comment out this define if you don't want WCECOMPAT and OpenSSL.
   Required if you compile for the emulator target because currently
   WCECOMPAT fails to compile after setting up the environment with
   wceemulator.bat */
#define WCECOMPAT_OPENSSL 1

#include <wincrypt.h>
#include <shlobj.h>
#include <stdlib.h>
#include <winbase.h>
#ifdef WCECOMPAT_OPENSSL
#include <openssl/evp.h>
#include <openssl/pkcs12.h>
#include <openssl/pem.h>
#include <openssl/x509.h>
#include <openssl/err.h>
#endif

#pragma comment(lib, "ceshell.lib")
#pragma comment(lib, "crypt32.lib")
#ifdef WCECOMPAT_OPENSSL
#pragma comment(lib, "libeay32.lib")
#pragma comment(lib, "wcecompat.lib")
#endif

#define MAXSTRLEN 256
/* The following parameters are for importing the private key.
   Change these if you have a smart card or other exotic requirements */
#define PROVIDER_NAME NULL
#define PROVIDER_TYPE PROV_RSA_FULL
/* "AT_KEYEXCHANGE: An exchange key pair used to encrypt session keys so
    that they can be safely stored and exchanged with other users.  
    AT_SIGNATURE:   A signature key pair used for authenticating digitally
    signed messages or files."
    This distinction was probably created due to U.S. export regulations,
    back when that was needed. */
#define PROVIDER_KEYSPEC AT_KEYEXCHANGE
//#define PROVIDER_KEYSPEC AT_SIGNATURE

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/* Display an error message in a message box */
void _err(const wchar_t* fcn, DWORD err = GetLastError()) 
{
	wchar_t szErrMsg[MAXSTRLEN]=L"";
	wchar_t szMsg[512]=L"";
	if ( FormatMessage( FORMAT_MESSAGE_FROM_SYSTEM, 0, err, 0, szErrMsg, sizeof szErrMsg / sizeof *szErrMsg, 0 ) )
	{
		 swprintf( szMsg, L"%s failed: %s", fcn, szErrMsg );
	} else 
	{
		swprintf( szMsg, L"%s failed: 0x%08X", fcn, err );
	}
	AfxMessageBox(szMsg, MB_ICONSTOP);
}

/* Display a (non-Microsoft) error message in a message box */
void _err2(const wchar_t* msg) 
{
	AfxMessageBox(msg, MB_ICONSTOP);
}

/* Determine if the certificate has the "CA:TRUE" X.509 basic constraint,
   i.e. if it is a Root or CA certificate. */
BOOL CertIsCACert(PCCERT_CONTEXT pCtx) 
{
	PCERT_EXTENSION pCertExt = NULL;
	BOOL fCA = FALSE; 
	PCERT_BASIC_CONSTRAINTS2_INFO pInfo = NULL;
	DWORD cbInfo = 0;

	pCertExt = CertFindExtension(	szOID_BASIC_CONSTRAINTS2,
									pCtx->pCertInfo->cExtension,
									pCtx->pCertInfo->rgExtension
								);
	if (pCertExt == NULL) 
	{
		// If no extension found, we assume it is a personal cert
		return FALSE;
	}

	if (!CryptDecodeObjectEx(	X509_ASN_ENCODING,
								X509_BASIC_CONSTRAINTS2,
								pCertExt->Value.pbData,
								pCertExt->Value.cbData,
								CRYPT_DECODE_ALLOC_FLAG,
								(PCRYPT_DECODE_PARA)NULL,
								&pInfo,
								&cbInfo))
	{
		/* Cannot decode certificate extension.
		   XXX Then we decide it is a personal personal cert
		   instead of exiting the program with _err(L"Cannot decode"); */
		return FALSE;
	} 

	if (pInfo)
	{
		fCA = pInfo->fCA;
		LocalFree(pInfo);
	}
	return fCA; 
}

/* Determine if the certificate has is self-signed,
   i.e. if the subject equals the issuer. Then it is a Root certificate. */
BOOL CertIsSelfsigned(PCCERT_CONTEXT pCtx)
{
	DWORD dwFlags = CERT_STORE_SIGNATURE_FLAG;
	
	if (!(CertCompareCertificateName(	X509_ASN_ENCODING,
										&pCtx->pCertInfo->Issuer,
										&pCtx->pCertInfo->Subject)))	
	{
		// wprintf(L"Not self-signed: Issuer != Subject\n");
		return FALSE;
	}

	if  (!(CertVerifySubjectCertificateContext(	pCtx, pCtx, &dwFlags)))
	{
		// wprintf(L"Not self-signed: CertVerifySubjectCertificateContext\n");
        return FALSE;
	}

	if (dwFlags != 0)
	{
		// wprintf(L"Not self-signed: dwflags\n");
		return FALSE;
	}

    return TRUE; 
}


/////////////////////////////////////////////////////////////////////////////
// CP12imprtDlg dialog

CP12imprtDlg::CP12imprtDlg(CWnd* pParent /*=NULL*/)
	: CDialog(CP12imprtDlg::IDD, pParent)
{

	TCHAR cert[MAX_PATH]=L"";

	//{{AFX_DATA_INIT(CP12imprtDlg)
	m_CertificateFile = L"";
	m_Password = L"";
	//}}AFX_DATA_INIT
	// Note that LoadIcon does not require a subsequent DestroyIcon in Win32
	m_hIcon = AfxGetApp()->LoadIcon(IDR_MAINFRAME);


	// By default, we will look for the default PKCS#12 file
	// in the "My Documents" folder.
	//
	// Find "My Documents" folder (actual name depends on PPC language version).
	if (!SHGetDocumentsFolder(L"\\",cert))
	{
		// Fallback to the English default
		_tcscpy(cert, L"\\My Documents");		
	}
	// Append slash and default filename.
	_tcscat(cert, L"\\user.pfx");
	// Use this file and path as the default
	m_CertificateFile = cert;
}

void CP12imprtDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CP12imprtDlg)
	DDX_Text(pDX, IDC_EDIT1, m_CertificateFile);
	DDX_Text(pDX, IDC_EDIT2, m_Password);
	//}}AFX_DATA_MAP
}

BEGIN_MESSAGE_MAP(CP12imprtDlg, CDialog)
	//{{AFX_MSG_MAP(CP12imprtDlg)
	ON_BN_CLICKED(IDC_BUTTON1, OnButton1)
	ON_EN_CHANGE(IDC_EDIT1, OnChangeEdit1)
	ON_EN_CHANGE(IDC_EDIT2, OnChangeEdit2)
	ON_BN_CLICKED(IDC_BUTTON2, OnButton2)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CP12imprtDlg message handlers

BOOL CP12imprtDlg::OnInitDialog()
{
	CDialog::OnInitDialog();

	// Set the icon for this dialog.  The framework does this automatically
	//  when the application's main window is not a dialog
	SetIcon(m_hIcon, TRUE);			// Set big icon
	SetIcon(m_hIcon, FALSE);		// Set small icon
	
	CenterWindow(GetDesktopWindow());	// center to the hpc screen

	// TODO: Add extra initialization here
	
	return TRUE;  // return TRUE  unless you set the focus to a control
}


void CP12imprtDlg::OnChangeEdit1() 
{
	// TODO: If this is a RICHEDIT control, the control will not
	// send this notification unless you override the CDialog::OnInitDialog()
	// function and call CRichEditCtrl().SetEventMask()
	// with the ENM_CHANGE flag ORed into the mask.
	
	// Editbox filename
	
}

void CP12imprtDlg::OnChangeEdit2() 
{
	// TODO: If this is a RICHEDIT control, the control will not
	// send this notification unless you override the CDialog::OnInitDialog()
	// function and call CRichEditCtrl().SetEventMask()
	// with the ENM_CHANGE flag ORed into the mask.
	
	// Editbox password
	
}

void CP12imprtDlg::OnButton2() 
{
	// TODO: Add your control notification handler code here
	//Browse button
	UpdateData(true);
	CFileDialog dlg(TRUE,
					NULL,
					NULL,
					OFN_HIDEREADONLY,
					L"Personal certs (*.p12;*.pfx)|*.p12;*.pfx|All files (*.*)|*.*||",
					NULL);

	// If the user selected a file, put the filename into the editbox.
	if(dlg.DoModal() == IDOK)
	{
		m_CertificateFile = dlg.GetPathName();
		UpdateData(false);
	}	
}

void CP12imprtDlg::OnButton1() 
{
	// Import button
	wchar_t name[MAXSTRLEN] = L"";
	wchar_t szMsg[MAXSTRLEN] = L"";
	wchar_t certType[MAXSTRLEN]= L"";

	HCERTSTORE myStore = 0;
	HCERTSTORE rootStore = 0;
	HCERTSTORE CAStore = 0;
	HCERTSTORE thisCertStore = 0;
	bool exportable = false;
	HCRYPTPROV hProv = 0;
	HCRYPTKEY hKey = 0;
	PCCERT_CONTEXT pctx = 0;
	BYTE *pbKeyBlob = NULL;	
	CRYPT_KEY_PROV_INFO KeyProvInfo;
	PCCERT_CONTEXT pctx2 = 0;
	PCCERT_CONTEXT pctx_ca = 0;
	HANDLE hfile = INVALID_HANDLE_VALUE;
	void* pfx = NULL;
	HANDLE hsection = 0;
	DWORD pfxfilesize = 0;

	char password[MAXSTRLEN];
	wchar_t containername[MAXSTRLEN]=L"p12imprt";
	unsigned char *pubkey=NULL;
	long privkeyblobsize = 0;
	char *privkeyblobptr = NULL;
	int res = 0;
	int nChoice = 0;
	// PKCS#12 file may contain multiple certificates.
	// These vars keep count of the numbers that were imported.
	// (Note that OpenSSL supports only 1 Personal cert per PKCS#12 file).
	int nImportedMYcerts = 0;
	int nImportedRootcerts = 0;
	int nImportedCAcerts = 0;
	int nSkippedcerts = 0;
	// Keep count of the type of certificate that was added in an iteration.
	int nRoottoimport = 0;
	int nCAtoimport = 0;
	int nMYtoimport = 0;
	DWORD importFlags = 0;
#ifdef WCECOMPAT_OPENSSL
	PKCS12 *p12=NULL;
	EVP_PKEY *pkey = NULL;
	X509 *pcert = NULL;
	STACK_OF(X509) *ca = NULL; 
	BIO *privkeyblobmem = NULL;
	BIO *pfxbio=NULL;
	int cacertnr = 0;
	int totalcacerts = 0;
#else
	FILE *fp=NULL;
#endif

	// Get the password and the filename from the dialog
	UpdateData(true);
	// We store each private key under a unique containername.
	// This way we can store multiple certificates in a cert store.
	swprintf(containername, L"p12imprt-%08X", Random());
	memset(password, 0, MAXSTRLEN);
	wcstombs(password, m_Password, MAXSTRLEN-1);
	// The password in the edit box is now no longer needed so it can be wiped.
	m_Password.Empty();

	// Open the MY, Root and CA cert stores.	
	myStore = CertOpenStore(	CERT_STORE_PROV_SYSTEM, 
								0, 
								0, 
								CERT_SYSTEM_STORE_CURRENT_USER, 
								L"MY");
	if (!myStore) 
	{
		_err(L"CertOpenStore MY");
		goto cleanup;
	}
	//CERT_SYSTEM_STORE_LOCAL_MACHINE
	rootStore = CertOpenStore(	CERT_STORE_PROV_SYSTEM, 
								0, 
								0, 
								CERT_SYSTEM_STORE_CURRENT_USER, 
								L"Root");

	if (!rootStore ) 
	{
		_err(L"CertOpenStore Root (locked? Smartphone?)");
	}

	CAStore = CertOpenStore(	CERT_STORE_PROV_SYSTEM, 
								0, 
								0, 
								CERT_SYSTEM_STORE_CURRENT_USER, 
								L"CA");

	if ((!CAStore ) && (rootStore))
	{
		_err(L"CertOpenStore CA (locked? Smartphone?)");
	}

#ifdef WCECOMPAT_OPENSSL
	OpenSSL_add_all_algorithms();
	ERR_load_crypto_strings();

	// Read PKCS#12 file
	hfile = CreateFileForMapping (m_CertificateFile,   // Open selected PKCS#12 file
                     GENERIC_READ,           // Open for reading
                     FILE_SHARE_READ,        // Share for reading
                     NULL,                   // No security
                     OPEN_EXISTING,          // Existing file only
                     FILE_ATTRIBUTE_NORMAL,  // Normal file
                     NULL);                  // No template file

	if (hfile == INVALID_HANDLE_VALUE) 
	{
		_err(L"CreateFileForMapping");
		goto cleanup;
	}

	hsection = CreateFileMapping(	hfile, 
									0, 
									PAGE_READONLY, 
									0, 
									0, 
									0);
	if (!hsection) 
	{
		_err(L"CreateFileMapping");
		goto cleanup;
	}

	pfx = MapViewOfFile(	hsection, 
							FILE_MAP_READ, 
							0, 
							0, 
							0);
	if (!pfx) 
	{
		_err(L"MapViewOfFile");
		goto cleanup;
	}
	privkeyblobmem = BIO_new(BIO_s_mem());
	if (privkeyblobmem == NULL)
	{
		_err2(L"Cannot create private key BIO");
		goto cleanup;
	}
	pfxfilesize = GetFileSize(hfile, NULL);
	if (pfxfilesize < 0)
	{
		_err(L"GetFileSize");
		goto cleanup;
	}
	pfxbio = BIO_new_mem_buf((void *)pfx, (int)pfxfilesize);
	if (pfxfilesize == NULL)
	{
		_err2(L"Cannot create BIO_new_mem_buf");
		goto cleanup;
	}
	p12=d2i_PKCS12_bio(pfxbio, NULL);
	if (!p12)
	{
		_err2(L"Not a valid P12 file");
		goto cleanup;
	}

	pkey = EVP_PKEY_new();
	if (pkey == NULL)
	{
		_err2(L"Cannot allocate memory for EVP_PKEY_new");
		goto cleanup;
	}
	pcert = X509_new();
	if (pcert == NULL)
	{
		_err2(L"Cannot allocate memory for X509_new");
		goto cleanup;
	}
	res=PKCS12_parse(p12, password, &pkey, &pcert, &ca);	
	if (res==0)
	{
		_err2(L"Cannot parse PKCS#12 file (incorrect password?)");
		goto cleanup;
	}
	res=i2b_PrivateKey_bio(privkeyblobmem, pkey);
	if (res < 0)
	{
		_err2(L"Cannot convert private key to privkeyblob");
		goto cleanup;
	}
	res=i2d_X509(pcert, &pubkey);
	if (res < 0)
	{
		_err2(L"Cannot convert cert to DER");
		goto cleanup;
	}
#else
	// If the WCECOMPAT and OpenSSL libraries are unavailable
	// (at the moment they don't compile for the emulator target),
	// read back a KEYBLOB from a file.
	// This is mainly for testing purposes.
	fp = fopen("\\Storage Card\\user.privatekeyblob","rb");
	if (fp == NULL)
	{
		_err2(L"Could not open privatekeyblob");
		goto cleanup;
	}
	// XXX This is a hack anyway so a fixed size is fine.
	char privkeyblob[10000];
	privkeyblobptr = privkeyblob;
	privkeyblobsize = (long) fread(privkeyblob, 1, 9999, fp);
	if (fp) fclose(fp);
	fp=NULL;

	// Now read a DER cert from a file. Testing only.
	fp = fopen("\\Storage Card\\user.publickeyblob","rb");
	if (fp == NULL)
	{
		_err2(L"Could not open publickeyblob");
		goto cleanup;
	}
	// XXX This is a hack anyway so a fixed size is fine.
	unsigned char pubkeyblob[10000];
	pubkey = pubkeyblob;
	res = (int) fread(pubkeyblob, 1, 9999, fp);
	if (fp) fclose(fp);
	fp=NULL;
#endif
	
	/* Create a MSCAPI cert context from this DER encoded cert */
	pctx = CertCreateCertificateContext(	X509_ASN_ENCODING,
											(BYTE*)pubkey,
											(DWORD)res);
	if (pctx == NULL)
	{
		_err(L"CertCreateCertificateContext");
		goto cleanup;
	}

	/* See if an equivalent cert already exists */
  	pctx2 = CertFindCertificateInStore(	myStore,
										X509_ASN_ENCODING,
										0,
										CERT_FIND_EXISTING,
										pctx,
										NULL);
	if (pctx2 != NULL)
	{
		// Cert already exists: ask user action
		// First, get the name in the cert
		if (CertGetNameString(pctx, CERT_NAME_FRIENDLY_DISPLAY_TYPE, 0, 0, name, sizeof name / sizeof *name)) 
		{
			// wprintf(L"Found a certificate in the PFX file: %s\n", name);
		} else 
		{
			_err(L"CertGetNameString");
			// XXX Do we have to exit?
		}

		swprintf( szMsg, L"An equivalent Personal cert already exists for '%s'. Overwrite?", name);
		nChoice = AfxMessageBox(szMsg, MB_YESNOCANCEL | MB_ICONQUESTION );
		if (nChoice == IDCANCEL)
		{
			goto cleanup2;
		}
		if (nChoice == IDNO)
		{
			goto addCA;
		}
	}

	if (exportable) 
	{
		importFlags |= CRYPT_EXPORTABLE;
	}
	
	// Get a handle to the default provider.
	if(!CryptAcquireContext	(			&hProv,
										containername,
										PROVIDER_NAME,
										PROVIDER_TYPE,
										importFlags))
	{
		// If it does not exist we create it
		if(GetLastError() == NTE_BAD_KEYSET)
		{
			if(!CryptAcquireContext(	&hProv,
										containername,
										PROVIDER_NAME,
										PROVIDER_TYPE,
										importFlags | CRYPT_NEWKEYSET))
			{
				_err(L"CryptAcquireContext CRYPT_NEWKEYSET");
				goto cleanup;
			}
		} else {
			_err(L"CryptAcquireContext (existing CSP)");
			goto cleanup;
		}
	}
	if (hProv == NULL)
	{
		_err(L"CryptAcquireContext is NULL,");
		goto cleanup;
	}

#ifdef WCECOMPAT_OPENSSL
	// Get pointer to key BLOB and size of key BLOB. Then import into CSP.
	privkeyblobsize = BIO_get_mem_data(privkeyblobmem, &privkeyblobptr);
#else
	// Get pointer to file data and size of data
#endif

	// Associate the certificate to the private key
	KeyProvInfo.pwszContainerName = containername; 
	KeyProvInfo.pwszProvName = PROVIDER_NAME; 
	KeyProvInfo.dwProvType = PROVIDER_TYPE;
	KeyProvInfo.dwFlags = 0; 
	KeyProvInfo.cProvParam = 0;
	KeyProvInfo.rgProvParam = NULL;
	KeyProvInfo.dwKeySpec = PROVIDER_KEYSPEC; 
	if(!CertSetCertificateContextProperty(pctx, CERT_KEY_PROV_INFO_PROP_ID, 0, &KeyProvInfo))
	{
		_err(L"CertSetCertificateContextProperty");
		goto cleanup;
	}

	/* If the cert exists we can replace it because the user
	   has already given permission for that. */
	if(!CertAddCertificateContextToStore(	myStore,
											pctx,
											CERT_STORE_ADD_REPLACE_EXISTING,
											NULL))
	{
		_err(L"CertAddCertificateContextToStore");
		goto cleanup;
	}
	
	if(!CryptImportKey(	hProv,
					(BYTE *)privkeyblobptr,
					(DWORD)privkeyblobsize,
					0,
					importFlags,
					&hKey))
	{
		_err(L"CryptImportKey");
		goto cleanup;
	}

	nImportedMYcerts++;

addCA:
	/* Add the root/CA certs. If the CA cert is self-signed,
		it will be added to the Root store. Otherwise it is added
		to the CA store. */
	nRoottoimport = 0;
	nCAtoimport = 0;
#ifdef WCECOMPAT_OPENSSL
	pubkey=NULL;
	if (pcert)
	{
		X509_free(pcert);
		pcert=NULL;
	}
	if (pctx)
	{
		CertFreeCertificateContext(pctx);
		pctx=NULL;
	}
	if (pubkey) 
	{
		OPENSSL_free(pubkey);
		pubkey=NULL;
	}
	if (ca == NULL)
	{
		goto cleanup2;
	}
	totalcacerts = ca->num;
	for (cacertnr = 0; cacertnr < totalcacerts; cacertnr++)
	{
		nRoottoimport = 0;
		nCAtoimport = 0;
		thisCertStore = 0;
		pcert = sk_X509_pop(ca);
		res=i2d_X509(pcert, &pubkey);
		if (res < 0)
		{
			_err2(L"Cannot convert cert to DER");
			goto cleanup;
		}
		/* Create a MSCAPI cert context from this DER encoded cert */
		pctx = CertCreateCertificateContext(	X509_ASN_ENCODING,
												(BYTE*)pubkey,
												(DWORD)res);
		if (pctx == NULL)
		{
				_err(L"CertCreateCertificateContext");
				goto cleanup;
		}
		if (CertGetNameString(pctx, CERT_NAME_FRIENDLY_DISPLAY_TYPE, 0, 0, name, sizeof name / sizeof *name)) 
		{
			// wprintf(L"Found a certificate in the PFX file: %s\n", name);
		} else 
		{
			_err(L"CertGetNameString");
			// XXX Do we have to exit?
		}
		
		if (CertIsSelfsigned(pctx) == TRUE)
		{
				/* Hrmpf. I wanted to test for the CA Basic Constraint
				first and then test for self-signedness, but joints like
				ValiCert don't even have the CA Basic Constraint in their
				root cert... */
				thisCertStore = rootStore;
				nRoottoimport = 1;
				wcscpy(certType, L"Root");				
		} else
		{
				if (CertIsCACert(pctx) == TRUE)
				{
					thisCertStore = CAStore;
					nCAtoimport = 1;
					wcscpy(certType, L"CA");
				}
				else
				{
					/* Not self-signed, no CA basic constraint:
					what is this cert doing in this list?!
					Must be a personal cert without private key?
					Skip it! */
					nSkippedcerts++;
					goto skipit;
				}
		}
		
		if (!thisCertStore)
		{
			nSkippedcerts++;
			goto skipit;
		}

		if (CertAddCertificateContextToStore(thisCertStore, pctx, CERT_STORE_ADD_NEW, 0)) 
		{
			// wprintf(L"Import of this certificate succeeded.\n");
			nImportedCAcerts += nCAtoimport;
			nImportedRootcerts += nRoottoimport;
		} else 
		{
			DWORD err = GetLastError();
			if (CRYPT_E_EXISTS == err) 
			{
				swprintf( szMsg, L"An equivalent %s cert already exists for '%s'. Overwrite?", certType, name);
				nChoice = AfxMessageBox(szMsg, MB_YESNOCANCEL | MB_ICONQUESTION );
				if (nChoice == IDCANCEL)
				{
					goto cleanup2;
				}
				if (nChoice == IDYES)
				{
					if (CertAddCertificateContextToStore(thisCertStore, pctx, CERT_STORE_ADD_REPLACE_EXISTING, 0)) 
					{
						//wprintf(L"Import of this certificate succeeded.\n");
						nImportedCAcerts += nCAtoimport;
						nImportedRootcerts += nRoottoimport;
					} else 
					{
						_err(L"CertAddCertificateContextToStore CERT_STORE_ADD_REPLACE_EXISTING");
						goto cleanup2;
					}
				} else 
				{
					// wprintf(L"No? OK, skipped.\n");
				}
			} else 
			{
				_err(L"CertAddCertificateContextToStore totally");
				goto cleanup2;
			}
		}

skipit:
		if (pcert)
		{
			X509_free(pcert);
			pcert=NULL;
		}
		if (pctx)
		{
			CertFreeCertificateContext(pctx);
			pctx=NULL;
		}
		if (pubkey) 
		{
			OPENSSL_free(pubkey);
			pubkey=NULL;
		}
	} /* End Root/CA loop */
#endif

cleanup2:
	// A graceful exit.
	if ((nImportedMYcerts + nImportedRootcerts + nImportedCAcerts) == 0)
	{
		AfxMessageBox(L"No certificates imported.", MB_ICONINFORMATION);
	} else
	{
		swprintf( szMsg, L"Personal certs: %d\nRoot certs: %d\nIntermed. certs: %d\nSkipped: %d\nTotal imported: %d", nImportedMYcerts, nImportedRootcerts, nImportedCAcerts, nSkippedcerts, nImportedMYcerts + nImportedRootcerts + nImportedCAcerts);
		AfxMessageBox(szMsg, MB_ICONINFORMATION);
	}

cleanup:
#ifdef WCECOMPAT_OPENSSL
	if (pkey)
	{
		EVP_PKEY_free(pkey);
	}
	if (pcert)
	{
		X509_free(pcert);
	}

	if (pubkey) 
	{
		OPENSSL_free(pubkey);
	}
	if (privkeyblobmem)
	{
		BIO_free_all(privkeyblobmem);
	}
	if (p12)
	{
		PKCS12_free(p12);
	}
	ERR_free_strings();
#else
	if (fp)
	{
		fclose(fp);
	}
#endif
	if (pctx)
	{
		CertFreeCertificateContext(pctx);
	}
		if (hProv)
	{
		CryptReleaseContext(hProv, 0);
	}
	if (pfx) 
	{
		UnmapViewOfFile(pfx);
	}
	if (hsection) 
	{
		CloseHandle(hsection);
	}
	if (hfile != INVALID_HANDLE_VALUE) 
	{
		CloseHandle(hfile);
	}
	if (myStore) 
	{
		CertCloseStore(myStore, 0);
	}
	if (rootStore) 
	{
		CertCloseStore(rootStore, 0);
	}
	if (CAStore) 
	{
		CertCloseStore(CAStore, 0);
	}
	// Set (wipe) the password in the dialog window
	UpdateData(false);

}
